// Copyright © 2010 Loïc Molinari <loicm@loicm.com>
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// Notes on threading, GLib mainloops and the rendering APIs:
//
//   Deface uses OpenGL for rendering its main window and GTK+ for displaying
// popups and menus. In order to benefit from all the goodness already done in
// GTK+ to communicate with the X11 server, the main window is created and
// handled by GDK. The idea is to create a GDK window and render it with OpenGL.
// One of the requirements for Deface is to be highly responsive reacting as
// fast as possible to user input events. That's why the application has been
// separated into several threads, with two of them dealing with rendering. The
// main thread runs the GMainLoop using the default GMainContext, GTK+ events
// and signals for popups and menus are dispatched from that thread. The GDK
// window rendered with OpenGL uses a dedicated display and runs in a dedicated
// thread which is not running a GMainLoop. It's a simple while(!quit) loop that
// sequentially dispatches events and renders the UI, blocking on the VBlank
// interrupt at the buffer-swap in order to throttle the thread.
//
//   GDK Events for that window are not dispatch the classic way, usually
// consisting to set a callback with gdk_event_handler_set() to the GSource,
// which in turn is polling the display connection file descriptor. This GSource
// is automatically attached to the default GMainContext by GDK when opening a
// new display. There are three problems with it:
//
// 1) callback's gonna be randomly called from the main thread
// 2) other GSources in the GMainContext can potentially delay event processing
// 3) plugs internal event handlers we don't care about (Expose event)
//
//   For these reasons, Deface simply removes that GSource from the default
// GMainContext and, at each iteration, directly gets from the display all the
// events queued so far using gdk_display_get_event(). That maybe sounds like
// *evil* for GTK+ gurus, but AFAICT I've not encountered a single problem using
// that technique. Feedback is really welcome about that subject anyway…

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif  // HAVE_CONFIG_H

#include "deface.h"
#include "deface_log.h"
#include <glib/gi18n.h>

static gboolean quit_cb (Deface* deface) {
  DEFACE_DASSERT (g_thread_self () == deface->main_thread);
  deface_ui_join_thread (deface);
  deface_camera_join_thread (deface);
  gtk_main_quit ();
  // Return FALSE to remove the source from the GMainContext (for the sake of
  // returning something cause the loop is stopped now anyway).
  return FALSE;
}

void deface_quit (Deface* deface, const gchar* error_message) {
  // This function can be called from several threads. We atomically set the
  // error message if any, the error message of the last caller will be
  // displayed at exit.
  if (error_message) {
    DEFACE_LOG (DEFACE_ERROR, "%s", error_message);
    g_atomic_pointer_set (&deface->error_message, (gpointer) error_message);
  }
  // The goal is to join all the threads of the application. In order to avoid
  // deadlocking we add a source to the default GMainContext, the callback's
  // gonna be dispatched in the main thread.
  g_idle_add ((GSourceFunc) quit_cb, deface);
}

int main (int argc, char* argv[]) {
  bindtextdomain (GETTEXT_PACKAGE, DEFACE_LOCALE_DIR);
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);
  g_set_application_name (PACKAGE_NAME);

  // Command-line initialization.
  GError* error = NULL;
  gboolean version_option = FALSE;
  gboolean argb_visual_option = TRUE;
  gboolean log_colored_option = TRUE;
  const gchar* log_threshold_option = "";
  const GOptionEntry entries[] = {
    { "version", 0, 0, G_OPTION_ARG_NONE, &version_option,
      "Print Deface version", NULL },
    { "no-argb-visual", 0, G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE,
      &argb_visual_option, "Disable use of X11 ARGB visual", NULL },
    { "log-no-color", 0, G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE,
      &log_colored_option, "Disable colored logging", NULL },
    { "log-threshold", 0, 0, G_OPTION_ARG_STRING, &log_threshold_option,
      "Set logging level threshold", "[info,warn,error]" },
    { NULL }
  };
  GOptionContext* options = g_option_context_new (
      "Apply effects on your face using a webcam");
  g_option_context_set_translation_domain (options, GETTEXT_PACKAGE);
  g_option_context_add_main_entries (options, entries, GETTEXT_PACKAGE);
  if (g_option_context_parse (options, &argc, &argv, &error) == FALSE) {
    fprintf (stderr, "Deface can't parse the options: %s\n", error->message);
    g_error_free (error);
    g_option_context_free (options);
    return 1;
  }
  g_option_context_free (options);

  if (version_option == TRUE) {
    fprintf (stdout, PACKAGE_NAME " " VERSION "\n");
    return 0;

  } else {
    // Deface global data structure is stored in the .bss section of the
    // executable (allocated when loaded). It's filled with NUL bytes by the
    // loader on Linux, we don't care for now but that might not be the case on
    // other systems.
    static Deface deface;

    // GTK+ initializes GDK which opens up a display that's called the default
    // display. This one's gonna be used for GTK+ popups and menus. For the
    // main window rendered in a dedicated thread, we open up another display
    // and handle events slightly differently than the usual way. See the notes
    // at the beginning of this file for more informations. We remove the
    // GSource which is handling events for that second display, based on the
    // fact that it's the second GSource added to the default GMainContext. It
    // really has to be done before the GMainLoop is launched, avoiding
    // potential deadlocks.
    g_thread_init (NULL);
    gdk_threads_init ();
    gtk_init (&argc, &argv);
    deface.ui.display = gdk_display_open (NULL);
    deface.ui.screen = gdk_display_get_screen (deface.ui.display, 0);
    // FIXME(loicm): Find a better way to remove that source cause a simple
    //     change in GDK could break our assumption.
    g_source_destroy (g_main_context_find_source_by_id (NULL, 2));

    // Logging initialization. Use INFO as threshold level for debug builds and
    // WARN for release builds.
    DefaceLogLevel threshold;
    if (!strncmp (log_threshold_option, "info", 4)) {
      threshold = DEFACE_INFO;
    } else if (!strncmp (log_threshold_option, "warn", 4)) {
      threshold = DEFACE_WARN;
    } else if (!strncmp (log_threshold_option, "error", 5)) {
      threshold = DEFACE_ERROR;
    } else {
#ifndef NDEBUG
      threshold = DEFACE_INFO;
#else
      threshold = DEFACE_WARN;
#endif
    }
    deface_log_init (threshold, log_colored_option);
    DEFACE_LOG (DEFACE_INFO, PACKAGE_NAME " " VERSION);

    // Store general informations.
#ifndef NDEBUG
    deface.main_thread = g_thread_self ();
#endif  // NDEBUG
    deface.argb_visual = argb_visual_option;
    deface.ui.geometry.width = 400.0f;
    deface.ui.geometry.height = 400.0f;

    // Initialize Deface components and start the GMainLoop using the default
    // GMainContext through the GTK+ main entry point.
    if (deface_ui_spawn_thread (&deface)) {
      if (deface_camera_spawn_thread (&deface)) {
        gdk_threads_enter ();
        gtk_main ();
        gdk_threads_leave ();
      } else {
        deface_ui_join_thread (&deface);
      }
    }

    // Leave the main thread displaying a popup in case of error.
    if (deface.error_message == NULL) {
      DEFACE_DLOG (DEFACE_INFO, "leaving main thread with error value 0");
      return 0;
    } else {
      // No need to take the GDK lock here since all threads have been joined.
      GtkWidget* dialog = gtk_message_dialog_new (
          NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Deface error");
      gtk_message_dialog_format_secondary_text (
          GTK_MESSAGE_DIALOG (dialog), "%s", deface.error_message);
      gtk_dialog_run (GTK_DIALOG (dialog));
      gtk_widget_destroy (dialog);
      DEFACE_DLOG (DEFACE_INFO, "leaving main thread with error value 1");
      return 1;
    }
  }
}
