// Copyright © 2010 Loïc Molinari <loicm@loicm.com>
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif  // HAVE_CONFIG_H

#include "deface_log.h"
#include <stdio.h>
#include <wchar.h>

gint32 _g_deface_log_threshold;
static gint32 g_colored;
static GTimeVal g_basetime;

static gchar* get_time (void) {
  // Get the diff.
  GTimeVal current, diff;
  g_get_current_time (&current);
  diff.tv_usec = current.tv_usec - g_basetime.tv_usec;
  diff.tv_sec = current.tv_sec - g_basetime.tv_sec;
  if (diff.tv_usec < 0) {
    diff.tv_sec--;
    diff.tv_usec += 1000000;
  }
  // Build a string removing the day and hour fields if not reached yet.
  glong days = diff.tv_sec / 86400;
  if (days == 0) {
    glong hours = diff.tv_sec / 3600;
    if (hours == 0) {
      return g_strdup_printf ("%02ld:%02ld:%06ld", (diff.tv_sec / 60) % 60,
                              diff.tv_sec % 60, diff.tv_usec);
    } else {
      return g_strdup_printf ("%02ld:%02ld:%02ld:%06ld", hours,
                              (diff.tv_sec / 60) % 60, diff.tv_sec % 60,
                              diff.tv_usec);
    }
  } else {
    return g_strdup_printf ("%ld%02ld:%02ld:%02ld:%06ld", days,
                            (diff.tv_sec / 3600) % 24,
                            (diff.tv_sec / 60) % 60,
                            diff.tv_sec % 60, diff.tv_usec);
  }
}

// TODO(loicm): Might use locking in order to write messages atomically and
//     behave correctly when logging from multiple threads. Consider using a
//     simple mutex since file locking is kinda broken on Linux (mid 2010). Note
//     that since such a beast will introduce a huge amount of sync points,
//     slowing down a lot thread parallelization, it would be reasonable to
//     make it optional.
void _deface_log_message (DefaceLogLevel level, const gchar* function,
                          const gchar* file, gint32 line,
                          const gchar* format, ...) {
  static const gchar* message_prefix[2][3] = {
    { "INFO \"", "WARN \"", "ERROR \"" },
    { "\033[37;01m", "\033[33;01m", "\033[31;01m" }
  };
  static const gchar* message_suffix[2] = {
    "\"", "\033[00m"
  };
  gchar* message;
  gchar* time;
  va_list args;

  va_start (args, format);
  message = g_strdup_vprintf (format, args);
  va_end (args);
  time = get_time ();
  fprintf (stderr, "%s %s%s%s %s %s %d %p\n", time,
           message_prefix[g_colored][level], message, message_suffix[g_colored],
           function, file, line, g_thread_self ());
  g_free (message);
  g_free (time);
}

void deface_log_init (DefaceLogLevel threshold, gboolean colored) {
  static gboolean initialized = FALSE;
  if (initialized == FALSE) {
    _g_deface_log_threshold = threshold;
    g_colored = colored;
    g_get_current_time (&g_basetime);
    initialized = TRUE;
  }
}

void deface_log_set_threshold (DefaceLogLevel level) {
  g_atomic_int_set (&_g_deface_log_threshold, level);
}

DefaceLogLevel deface_log_get_threshold (void) {
  return g_atomic_int_get (&_g_deface_log_threshold);
}

void deface_log_set_colored (gboolean colored) {
  g_atomic_int_set (&g_colored, colored ? 1 : 0);
}

gboolean deface_log_get_colored (void) {
  return g_atomic_int_get (&g_colored) == 0 ? FALSE : TRUE;
}
