// Copyright © 2010 Loïc Molinari <loicm@loicm.com>
// Use of this source code is governed by a BSD-style license that can be found
// in the LICENSE file.

#ifndef DEFACE_LOG_H_
#define DEFACE_LOG_H_

#include <glib.h>

typedef enum {
  DEFACE_INFO,   // Information message.
  DEFACE_WARN,   // Warning message.
  DEFACE_ERROR,  // Error message.
} DefaceLogLevel;

// Implementation details, don't mess with these symbols.
void _deface_log_message (
    DefaceLogLevel level, const gchar* function, const gchar* file, gint32 line,
    const gchar* format, ...) G_GNUC_PRINTF (5, 6) G_GNUC_NO_INSTRUMENT;
extern gint32 _g_deface_log_threshold;

// Must be called before using logging facilities.
void deface_log_init (DefaceLogLevel threshold, gboolean colored);

// The log level threshold can be used to quickly discard log messages falling
// under. By default, the threshold is set to DEFACE_INFO for debug builds and
// to DEFACE_WARN for release builds (using the -DNDEBUG flag).
void deface_log_set_threshold (DefaceLogLevel level);
DefaceLogLevel deface_log_get_threshold (void);

// Log messages can be colored to improve readability. Most terminals supports
// colored texts using the Ecma-48 (ISO/IEC-6429) standard "Control
// functions for coded character sets" using escape sequences. By default, the
// log messages are colored.
void deface_log_set_colored (gboolean colored);
gboolean deface_log_get_colored (void);

// Log a message to stderr using the given level.
#define DEFACE_LOG(level,...) G_STMT_START {                                   \
  if (level >= g_atomic_int_get (&_g_deface_log_threshold))                    \
    _deface_log_message ((level), __func__, __FILE__, __LINE__, __VA_ARGS__);  \
} G_STMT_END

// Conditionally log a message to stderr using the given level.
#define DEFACE_LOG_IF(level,cond,...) G_STMT_START {                           \
  if (((level) >= g_atomic_int_get (&_g_deface_log_threshold)) && (cond))      \
    _deface_log_message ((level), __func__, __FILE__, __LINE__, __VA_ARGS__);  \
} G_STMT_END

// Log an error message and break the program execution if the assertion fails.
#define DEFACE_ASSERT(cond) G_STMT_START {                                     \
  if (G_UNLIKELY ((cond) == FALSE)) {                                          \
    _deface_log_message (DEFACE_ERROR, __func__, __FILE__, __LINE__,           \
                         "assertion `"#cond"' failed");                        \
    G_BREAKPOINT ();                                                           \
  }                                                                            \
} G_STMT_END

// Debug mode logging macros, compiled away to nothing for release builds (using
// the -DNDEBUG flag).
#ifndef NDEBUG
#define DEFACE_DLOG(level,...)                                                 \
  DEFACE_LOG ((level), __VA_ARGS__)
#define DEFACE_DLOG_IF(level,cond,...)                                         \
  DEFACE_LOG_IF ((level), (cond), __VA_ARGS__)
#define DEFACE_DASSERT(cond)                                                   \
  DEFACE_ASSERT ((cond))
#else
#define DEFACE_DLOG(level,...)
#define DEFACE_DLOG_IF(level,cond,...)
#define DEFACE_DASSERT(cond)
#endif

// Redefine the standard assert to use our log system.
#undef assert
#define assert(x) DEFACE_DASSERT(x)

#endif  // DEFACE_LOG_H_
