// Copyright © 2010 Loïc Molinari <loicm@loicm.com>
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// Deface effects are implemented using per-fragment GPU programmability. Since
// we want to run on a wide range of hardware, we require support for the
// multi-vendor OpenGL 1.x ARB_fragment_program extension which provides basic
// per-fragment programmability for the first generation of programmable GPUs
// (mid-2002). Support for per-vertex programmability is not mandatory as it's
// not as widely supported, so we keep using the fixed-function vertex
// pipeline. We could detect and use per-vertex programmability when it's
// available as it's recommended these days, but that implies maintaining one
// more code path and Deface is very light on vertex processing anyway. Program
// strings are embedded inside the read-only data section of the exe. A new line
// sequence ('\n') is inserted at the end of each instruction to make easier
// understanding compilation errors.

#ifndef DEFACE_PROGRAMS_H_
#define DEFACE_PROGRAMS_H_

#include <glib.h>

#define ARBFP_HEADER                      \
  "!!ARBfp1.0                         \n" \
  "OPTION ARB_precision_hint_fastest ;\n" \
  "OUTPUT out_color = result.color   ;\n"

#define ARBFP_FOOTER \
  "END"

// Glossy effect on the window frame.
// FIXME(loicm): Quick hack that needs to be cleaned up, tweaked and detailed.
static const gchar glossy_window[] =
    ARBFP_HEADER
    "PARAM  geometry = program.local[0]             ;\n"
    "PARAM  scale = { 0.001, 0.0015, 412.0, 412.0 } ;\n"
    "ATTRIB tc = fragment.texcoord[0]               ;\n"
    "ATTRIB in_color = fragment.color               ;\n"
    "TEMP   r0, r1                                  ;\n"
    "RCP r0.x, geometry.z             ;\n"
    "RCP r0.y, geometry.w             ;\n"
    "MAD r1, scale.zwzw, r0, tc       ;\n"
    "MAD r0.xy, scale, geometry, r1   ;\n"
    "TEX r1, r0, texture[0], 2D       ;\n"
    "MUL r0, r1, 0.6                  ;\n"
    "ADD out_color, in_color, r0.aaaa ;\n"
    "MOV out_color.a, in_color        ;\n"
    ARBFP_FOOTER;

// Set the fragment color to the primary color attribute.
static const gchar fragment_color[] =
    ARBFP_HEADER
    "ATTRIB in_color = fragment.color ;\n"
    "MOV out_color, in_color          ;\n"
    ARBFP_FOOTER;

// Render an anti-aliased arc filled with white pixels.
// FIXME(loicm): Hard-coded radius.
static const gchar filled_arc[] =
    ARBFP_HEADER
    "ATTRIB tc = fragment.texcoord[0] ;\n"
    "TEMP   r0, r1, r2, distance      ;\n"
    // Pythagoras's theorem gives us the distance to the center of the arc.
    "MUL r0, tc, tc     ;\n"
    "ADD r1, r0.x, r0.y ;\n"
    "RSQ r0, r1.x       ;\n"
    "RCP distance, r0.x ;\n"
    // Set pixels lying under the distance to [1,1,1,1] with a gradient
    // beginning at 15.5. We're extending the gradient till 16.25 as it seems
    // visually better.
    "SLT r0, distance, 16.25 ;\n"
    "SUB r1, distance, 15.5  ;\n"
    "MUL r2, r1, 1.5         ;\n"
    "SUB r1, 1.0, r2         ;\n"
    "MUL out_color, r0, r1   ;\n"
    ARBFP_FOOTER;

// Render two anti-aliased stroked arcs filled with colors given as parameters.
// FIXME(loicm): Hard-coded radius.
static const gchar stroked_arcs[] =
    ARBFP_HEADER
    "PARAM  inner_color = program.local[0]     ;\n"
    "PARAM  outer_color = program.local[1]     ;\n"
    "ATTRIB tc = fragment.texcoord[0]          ;\n"
    "TEMP   r0, r1, r2, distance, inner, outer ;\n"
    // Pythagoras's theorem gives us the distance to the center of the arc.
    "MUL r0, tc, tc     ;\n"
    "ADD r1, r0.x, r0.y ;\n"
    "RSQ r0, r1.x       ;\n"
    "RCP distance, r0.x ;\n"
    // Map all fragment in [14.5, 15.0] to [1.0, 0.0].
    "SUB r0, distance, 14.5 ;\n"
    "MAD r1, r0, -2.0, 1.0  ;\n"
    // Map all fragment in [14.0, 14.5] to [0.0, 1.0].
    "SUB r0, distance, 14.0 ;\n"
    "MUL r2, r0, 2.0        ;\n"
    // Store our anti-aliased inner stroked arc with its right color. The last
    // op is clamped to remove negative values that can break the final blend.
    "MUL r0, r1, r2                 ;\n"
    "MUL_SAT inner, r0, inner_color ;\n"
    // Map all fragment in [15.5, 16.0] to [1.0, 0.0]. As we do for the filled
    // arc, we're extending the gradient till 16.25 as it seems visually better.
    "SUB r0, distance, 15.5 ;\n"
    "MAD r1, r0, -1.5, 1.0  ;\n"
    // Map all fragment in [15.0, 15.5] to [0.0, 1.0].
    "SUB r0, distance, 15.0 ;\n"
    "MUL r2, r0, 2.0        ;\n"
    // Store our anti-aliased outer stroked arc with its right color. The last
    // op is clamped to remove negative values that can break the final blend.
    "MUL r0, r1, r2                 ;\n"
    "MUL_SAT outer, r0, outer_color ;\n"
    // Blend the outer stroked arc over the inner stroked one.
    "SUB r0, 1.0, outer.aaaa         ;\n"
    "MAD out_color, inner, r0, outer ;\n"
    ARBFP_FOOTER;

// 
static const gchar foo4[] =
    ARBFP_HEADER
    "MOV out_color, {0.0, 1.0, 0.0, 1.0};"
    "END";

// 
static const gchar foo5[] =
    ARBFP_HEADER
    "MOV out_color, {0.0, 1.0, 0.0, 1.0};"
    "END";

// 
static const gchar foo6[] =
    ARBFP_HEADER
    "MOV out_color, {0.0, 1.0, 0.0, 1.0};"
    "END";

// 
static const gchar foo7[] =
    ARBFP_HEADER
    "MOV out_color, {0.0, 1.0, 0.0, 1.0};"
    "END";

// 
static const gchar foo8[] =
    ARBFP_HEADER
    "MOV out_color, {0.0, 1.0, 0.0, 1.0};"
    "END";

// 
static const gchar foo9[] =
    ARBFP_HEADER
    "MOV out_color, {0.0, 1.0, 0.0, 1.0};"
    "END";

// Declare our programs with their respective lengths resolved at compile-time,
// whitout the null character terminator as it's required by glProgramStringARB.
static const struct {
  const gint32 length;
  const gchar* source;
} deface_program[] = {
  { sizeof (glossy_window) - 1, glossy_window },
  { sizeof (fragment_color) - 1, fragment_color },
  { sizeof (filled_arc) - 1, filled_arc },
  { sizeof (stroked_arcs) - 1, stroked_arcs },
  { sizeof (foo4) - 1, foo4 },
  { sizeof (foo5) - 1, foo5 },
  { sizeof (foo6) - 1, foo6 },
  { sizeof (foo7) - 1, foo7 },
  { sizeof (foo8) - 1, foo8 },
  { sizeof (foo9) - 1, foo9 }
};

// Program identifiers.
enum {
  DEFACE_PROGRAM_GLOSSY_WINDOW = 0,
  DEFACE_PROGRAM_FRAGMENT_COLOR,
  DEFACE_PROGRAM_FILLED_ARC,
  DEFACE_PROGRAM_STROKED_ARCS,
  DEFACE_PROGRAM_FOO4,
  DEFACE_PROGRAM_FOO5,
  DEFACE_PROGRAM_FOO6,
  DEFACE_PROGRAM_FOO7,
  DEFACE_PROGRAM_FOO8,
  DEFACE_PROGRAM_FOO9,
  DEFACE_NR_PROGRAMS
};

#endif  // DEFACE_PROGRAMS_H_
