// Copyright © 2010 Loïc Molinari <loicm@loicm.com>
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif  // HAVE_CONFIG_H

#include "deface.h"
#include "deface_log.h"
#include <X11/extensions/Xrender.h>
#include <gdk/gdkkeysyms.h>
#include <string.h>

// UI parameters constants.
#define MIN_WINDOW_WIDTH          400
#define MIN_WINDOW_HEIGHT         400
#define MIN_WINDOW_ASPECT         0.5
#define MAX_WINDOW_ASPECT         2.0
#define FRAME_INNER_BORDER_R      255
#define FRAME_INNER_BORDER_G      255
#define FRAME_INNER_BORDER_B      255
#define FRAME_OUTER_BORDER_R       0
#define FRAME_OUTER_BORDER_G       0
#define FRAME_OUTER_BORDER_B       0
#define FRAME_BORDER_TRANSLUCENCY 150
#define FRAME_TRANSLUCENCY        210
#define FACE_AREA_R               0
#define FACE_AREA_G               0
#define FACE_AREA_B               0
#define FACE_AREA_TRANSLUCENCY    250
#define CORNER_RADIUS             16

// Pack 4 bytes (uint8) in a dword (uint32). b0 is the least significant byte,
// b3 the most significant.
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
#define PACK_UINT8(b0,b1,b2,b3) ((b3)<<24 | (b2)<<16 | (b1)<<8 | (b0))
#else  // G_BYTE_ORDER == G_BIG_ENDIAN
#define PACK_UINT8(b0,b1,b2,b3) ((b0)<<24 | (b1)<<16 | (b2)<<8 | (b3))
#endif

// Set the color of a vertex given the R, G, B and A bytes.
#define SET_VERTEX_COLOR_UINT8(v,i,r,g,b,a) \
  ((guint8*)v)[i*4+0] = r;                  \
  ((guint8*)v)[i*4+1] = g;                  \
  ((guint8*)v)[i*4+2] = b;                  \
  ((guint8*)v)[i*4+3] = a

// Set the color of a vertex given a RGBA dword. R must be stored in the least
// significant byte, A in the most significant.
#define SET_VERTEX_COLOR_UINT32(v,i,c) \
  ((guint32*)v)[i] = c

// Set the position of a vertex.
#define SET_VERTEX_POSITION(v,i,x,y) \
  ((gfloat*)v)[i*2+0] = x;           \
  ((gfloat*)v)[i*2+1] = y

// Set the texture coordinates of a vertex.
#define SET_VERTEX_TEXCOORD(v,i,s,t) \
  ((gfloat*)v)[i*2+0] = s;           \
  ((gfloat*)v)[i*2+1] = t

// Set the indices of a triangle.
#define SET_TRIANGLE_INDICES(v,i,a,b,c) \
  ((guint16*)v)[i*3+0] = a;             \
  ((guint16*)v)[i*3+1] = b;             \
  ((guint16*)v)[i*3+2] = c

// Porter/Duff blending operators.
typedef enum {
  BLENDING_CLEAR    = 0,
  BLENDING_SRC      = 1,
  BLENDING_DST      = 2,
  BLENDING_SRC_OVER = 3,
  BLENDING_DST_OVER = 4,
  BLENDING_SRC_IN   = 5,
  BLENDING_DST_IN   = 6,
  BLENDING_SRC_OUT  = 7,
  BLENDING_DST_OUT  = 8,
  BLENDING_SRC_ATOP = 9,
  BLENDING_DST_ATOP = 10,
  BLENDING_XOR      = 11,
  BLENDING_PLUS     = 12
} BlendingOp;

// UI flags.
enum {
  FLAG_X11_ARGB_VISUAL   = (1 << 0),
  FLAG_GLX_FBCONFIG      = (1 << 1),
  FLAG_GLX_SWAP_CONTROL  = (1 << 2),
  FLAG_GLX_VIDEO_SYNC    = (1 << 3),
  FLAG_OPENGL_ARBVP      = (1 << 4),
  FLAG_OPENGL_ARBFP      = (1 << 5),
  FLAG_OPENGL_VBO        = (1 << 6),
  FLAG_OPENGL_PBO        = (1 << 7),
  FLAG_OPENGL_NPOT       = (1 << 8),
  FLAG_SCREEN_COMPOSITED = (1 << 9)
};

// UI error ids.
enum {
  ERROR_NO_GLX_EXTENSION         = 0,
  ERROR_CANNOT_QUERY_GLX_VERSION = 1,
  ERROR_BAD_GLX_VERSION          = 2,
  ERROR_NO_X11_VISUAL            = 3,
  ERROR_NO_GDK_VISUAL            = 4,
  ERROR_CANNOT_CREATE_CONTEXT    = 5,
  ERROR_CANNOT_BIND_CONTEXT      = 6,
  ERROR_OPENGL_ARBFP_REQUIRED    = 7,
  ERROR_OPENGL_ARBFP_BROKEN      = 8,
  ERROR_CANNOT_COMPILE_PROGRAM   = 9,
  ERROR_CANNOT_CREATE_TEXTURE    = 10
};

// UI error messages.
static const gchar* error_message[] = {
  "X11 server doesn't support the GLX extension.",
  "Can't query GLX version.",
  "GLX version 1.2+ required.",
  "Can't find a X11 visual.",
  "Can't find a GdkVisual.",
  "Can't create an OpenGL rendering context.",
  "Can't bind the OpenGL rendering context.",
  "OpenGL ARB_fragment_program extension required.",
  "OpenGL ARB_fragment_program extension broken.",
  "Can't compile a fragment program.",
  "Can't create a texture."
};

// Set the OpenGL blending state in terms of Porter/Duff binary compositing
// operators. The arithmetic specified by these operator requires both source
// and destination buffers to be filled with premultiplied pixels. Output pixels
// are premultiplied as well, which is actually what's required by the XRender
// specification for correct desktop window compositing, neat!
static inline void set_blending_state (BlendingOp op) {
  GLenum factors[13][2] = {
    { GL_ZERO,                GL_ZERO                },  // BLENDING_CLEAR
    { GL_ONE,                 GL_ZERO                },  // BLENDING_SRC
    { GL_ZERO,                GL_ONE                 },  // BLENDING_DST
    { GL_ONE,                 GL_ONE_MINUS_SRC_ALPHA },  // BLENDING_SRC_OVER
    { GL_ONE_MINUS_DST_ALPHA, GL_ONE                 },  // BLENDING_DST_OVER
    { GL_DST_ALPHA,           GL_ZERO                },  // BLENDING_SRC_IN
    { GL_ZERO,                GL_SRC_ALPHA           },  // BLENDING_DST_IN
    { GL_ONE_MINUS_DST_ALPHA, GL_ZERO                },  // BLENDING_SRC_OUT
    { GL_ZERO,                GL_ONE_MINUS_SRC_ALPHA },  // BLENDING_DST_OUT
    { GL_DST_ALPHA,           GL_ONE_MINUS_SRC_ALPHA },  // BLENDING_SRC_ATOP
    { GL_ONE_MINUS_DST_ALPHA, GL_SRC_ALPHA           },  // BLENDING_DST_ATOP
    { GL_ONE_MINUS_DST_ALPHA, GL_ONE_MINUS_SRC_ALPHA },  // BLENDING_XOR
    { GL_ONE,                 GL_ONE                 }   // BLENDING_PLUS
  };
  glBlendFunc (factors[op][0], factors[op][1]);
}

// Set the sampling state of the texture currently bound. OpenGL binds it to a
// texture object so it might just be called once (compared to Direct3D which
// binds it to the device).
static inline void set_sampling_state (GLenum target, GLint min_filter,
                                       GLint mag_filter, GLint wrap_s,
                                       GLint wrap_t) {
  glTexParameteri (target, GL_TEXTURE_MIN_FILTER, min_filter);
  glTexParameteri (target, GL_TEXTURE_MAG_FILTER, mag_filter);
  glTexParameteri (target, GL_TEXTURE_WRAP_S, wrap_s);
  glTexParameteri (target, GL_TEXTURE_WRAP_T, wrap_t);
}

// Checks whether an extension is supported by the GLX or OpenGL implementation
// given the extension name and the list of supported extensions.
static gboolean is_extension_supported (const gchar *extensions,
                                        const gchar *extension) {
  if (extensions != NULL && extension != NULL) {
    const gsize len = strlen (extension);
    gchar* p = (gchar*) extensions;
    gchar* end = p + strlen (p);
    while (p < end) {
      const gsize size = strcspn (p, " ");
      if (len == size && strncmp (extension, p, size) == 0)
        return TRUE;
      p += size + 1;
    }
  }
  return FALSE;
}

static gboolean init_x11 (Deface* deface) {
  DEFACE_DASSERT (g_thread_self () == deface->ui.thread);
  DefaceUI* ui = &deface->ui;
  ui->xdisplay = GDK_SCREEN_XDISPLAY (ui->screen);
  ui->xscreen_nr = GDK_SCREEN_XNUMBER (ui->screen);

  // Check GLX implementation.
  gint32 error, event;
  if (glXQueryExtension (ui->xdisplay, &error, &event)) {
    if (glXQueryVersion (ui->xdisplay, &ui->glx_major, &ui->glx_minor)) {
      if (ui->glx_major > 1 || (ui->glx_major == 1 && ui->glx_minor >= 2)) {
        GlxExtensions* glx = &ui->glx;

        // Search for supported GLX extensions.
        const struct { const gchar* name; const guint16 flag; } extension[] = {
          { "GLX_SGIX_fbconfig", FLAG_GLX_FBCONFIG },
          { "GLX_SGI_swap_control", FLAG_GLX_SWAP_CONTROL },
          { "GLX_SGI_video_sync", FLAG_GLX_VIDEO_SYNC },
          { NULL, 0 }
        };
        const gchar* extensions = glXQueryExtensionsString (
            ui->xdisplay, ui->xscreen_nr);
        guint16 flags = 0;
        for (gint32 i = 0; extension[i].name != NULL; i++)
          if (is_extension_supported (extensions, extension[i].name) == TRUE)
            flags |= extension[i].flag;

        // Load GLX_SGIX_fbconfig extension functions.
        if (flags & FLAG_GLX_FBCONFIG) {
          if (ui->glx_major == 1 && ui->glx_minor >= 3) {
            glx->choose_fb_config = (ChooseFBConfigFunc) glXGetProcAddressARB (
                (const GLubyte*) "glXChooseFBConfig");
            glx->get_visual_from_fb_config = (GetVisualFromFBConfigFunc)
                glXGetProcAddressARB ((const GLubyte*)
                                      "glXGetVisualFromFBConfig");
          } else {
            glx->choose_fb_config = (ChooseFBConfigFunc) glXGetProcAddressARB (
                (const GLubyte*) "glXChooseFBConfigSGIX");
            glx->get_visual_from_fb_config = (GetVisualFromFBConfigFunc)
                glXGetProcAddressARB ((const GLubyte*)
                                      "glXGetVisualFromFBConfigSGIX");
          }
          if (glx->choose_fb_config == NULL ||
              glx->get_visual_from_fb_config == NULL) {
            flags &= ~FLAG_GLX_FBCONFIG;
          }
        }

        // Load GLX_SGI_swap_control extension function.
        if (flags & FLAG_GLX_SWAP_CONTROL) {
          glx->swap_interval =
              (SwapIntervalFunc) glXGetProcAddressARB (
                  (const GLubyte*) "glXSwapIntervalSGI");
          if (glx->swap_interval == NULL)
            flags &= ~FLAG_GLX_SWAP_CONTROL;
        }

        // Load GLX_SGI_video_sync extension functions.
        if (flags & FLAG_GLX_VIDEO_SYNC) {
          glx->get_video_sync = (GetVideoSyncFunc) glXGetProcAddressARB (
              (const GLubyte*) "glXGetVideoSyncSGI");
          glx->wait_video_sync = (WaitVideoSyncFunc) glXGetProcAddressARB (
              (const GLubyte*) "glXWaitVideoSyncSGI");
          if (glx->get_video_sync == NULL || glx->wait_video_sync == NULL)
            flags &= ~FLAG_GLX_VIDEO_SYNC;
        }

        DEFACE_LOG (DEFACE_INFO, "GLX_SGIX_fbconfig: %s",
                    flags & FLAG_GLX_FBCONFIG ? "yes" : "no");
        DEFACE_LOG (DEFACE_INFO, "GLX_SGI_swap_control: %s",
                    flags & FLAG_GLX_SWAP_CONTROL ? "yes" : "no");
        DEFACE_LOG (DEFACE_INFO, "GLX_SGI_video_sync: %s",
                    flags & FLAG_GLX_VIDEO_SYNC ? "yes" : "no");
        ui->flags |= flags;

        return TRUE;
      } else {
        deface_quit (deface, error_message[ERROR_BAD_GLX_VERSION]);
      }
    } else {
      deface_quit (deface, error_message[ERROR_CANNOT_QUERY_GLX_VERSION]);
    }
  } else {
    deface_quit (deface, error_message[ERROR_NO_GLX_EXTENSION]);
  }
  return FALSE;
}

// If argb is TRUE, search for an ARGB visual falling back to RGB if not
// found. Store the retrieved xvisual or NULL in the data structure and set the
// ARGB_VISUAL flag if needed. Create the window using the retrieved visual.
static gboolean init_window (Deface* deface) {
  DEFACE_DASSERT (g_thread_self () == deface->ui.thread);
  DefaceUI* ui = &deface->ui;
  GlxExtensions* glx = &ui->glx;
  Display* xdisplay = ui->xdisplay;
  XVisualInfo* xvisual = NULL;
  guint8 flags = 0;

  if (ui->flags & FLAG_GLX_FBCONFIG) {
    // Get the available frame buffer configurations.
    const gint32 attribs[] = {
      GLX_DOUBLEBUFFER, True,
      GLX_RED_SIZE, 1,
      GLX_GREEN_SIZE, 1,
      GLX_BLUE_SIZE, 1,
      GLX_ALPHA_SIZE, 1,
      GLX_DEPTH_SIZE, 0,  // No depth buffer required.
      GLX_RENDER_TYPE, GLX_RGBA_BIT,
      GLX_DRAWABLE_TYPE, GLX_WINDOW_BIT,
      None
    };
    gint32 nr_configs;
    GLXFBConfig* fb_configs = glx->choose_fb_config (
        xdisplay, ui->xscreen_nr, attribs, &nr_configs);

    if (fb_configs != NULL) {
      gint32 event, error;
      if (deface->argb_visual == FALSE) {
        // No need for an ARGB visual, we use the first config retrieved.
        xvisual = glx->get_visual_from_fb_config (xdisplay, fb_configs[0]);
      } else if (XRenderQueryExtension (xdisplay, &event, &error) != 0) {
        // Search for an ARGB visual. No black magic here, the code comes from
        // NVIDIA and is shipped in the README file of their proprietary driver.
        gint32 i;
        for (i = 0; i < nr_configs; i++) {
          XRenderPictFormat* format;
          xvisual = glx->get_visual_from_fb_config (xdisplay, fb_configs[i]);
          if (xvisual == NULL)
            continue;
          format = XRenderFindVisualFormat (xdisplay, xvisual->visual);
          if (format == NULL) {
            XFree (xvisual);
            continue;
          }
          if (format->type == PictTypeDirect && format->direct.alphaMask > 0) {
            flags |= FLAG_X11_ARGB_VISUAL;
            break;
          }
          XFree (xvisual);
        }
        if (i == nr_configs) {
          // There's no ARGB visual, we use the first config retrieved.
          xvisual = glx->get_visual_from_fb_config (xdisplay, fb_configs[0]);
        }
      }
      XFree (fb_configs);
    }
  }

  if (xvisual == NULL) {
    // Last chance, using the old visual API. */
    gint attribs[] = {
      GLX_DOUBLEBUFFER, 1,
      GLX_RGBA,
      GLX_RED_SIZE, 1,
      GLX_GREEN_SIZE, 1,
      GLX_BLUE_SIZE, 1,
      GLX_DEPTH_SIZE, 0,  // No depth buffer required.
      None
    };
    xvisual = glXChooseVisual (xdisplay, ui->xscreen_nr, attribs);
  }

  if (xvisual != NULL) {
    gdk_threads_enter ();
    // Lookup the GdkVisual wrapping our X11 visual.
    GdkVisual* visual = gdk_x11_screen_lookup_visual (ui->screen,
                                                      xvisual->visualid);
    if (visual != NULL) {
      DEFACE_LOG (DEFACE_INFO, "loaded a %u-bit X11 visual (0x%2x)",
                  xvisual->depth, (gint32) xvisual->visualid);

      // Create the window with a constrained minimal size and without
      // decorations.
      GdkWindowAttr attribs;
      attribs.event_mask = GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
          GDK_POINTER_MOTION_MASK | GDK_KEY_PRESS_MASK | GDK_FOCUS_CHANGE_MASK |
          GDK_STRUCTURE_MASK;
      attribs.width = deface->ui.geometry.width;
      attribs.height = deface->ui.geometry.height;
      attribs.wclass = GDK_INPUT_OUTPUT;
      attribs.visual = visual;
      attribs.colormap = gdk_colormap_new (visual, FALSE);
      attribs.window_type = GDK_WINDOW_TOPLEVEL;
      attribs.type_hint = GDK_WINDOW_TYPE_HINT_NORMAL;
      ui->window = gdk_window_new (gdk_screen_get_root_window (ui->screen),
                                   &attribs, GDK_WA_VISUAL | GDK_WA_COLORMAP |
                                   GDK_WA_TYPE_HINT);
      const GdkGeometry geometry = {
        .min_width = MIN_WINDOW_WIDTH, .min_height = MIN_WINDOW_HEIGHT,
        .min_aspect = MIN_WINDOW_ASPECT, .max_aspect = MAX_WINDOW_ASPECT
      };
      gdk_window_set_geometry_hints (
          ui->window, &geometry, GDK_HINT_MIN_SIZE | GDK_HINT_ASPECT);
      gdk_window_set_decorations (ui->window, 0);
      ui->xwindow = GDK_WINDOW_XWINDOW (ui->window);
      ui->xvisual = xvisual;
      ui->flags |= flags;
      gdk_flush ();
      gdk_threads_leave ();

      return TRUE;
    } else {
      deface_quit (deface, error_message[ERROR_NO_GDK_VISUAL]);
    }
    gdk_threads_leave ();
  } else {
    deface_quit (deface, error_message[ERROR_NO_X11_VISUAL]);
  }
  return FALSE;
}

static gboolean init_opengl (Deface* deface) {
  DEFACE_DASSERT (g_thread_self () == deface->ui.thread);
  DefaceUI* ui = &deface->ui;

  // Create and bound the OpenGL context to the UI thread.
  ui->xcontext = glXCreateContext (ui->xdisplay, ui->xvisual, NULL, True);
  if (ui->xcontext) {
    if (glXMakeCurrent (ui->xdisplay, ui->xwindow, ui->xcontext)) {
      DEFACE_DLOG (DEFACE_INFO, "created %s rendering context",
                   glXIsDirect (ui->xdisplay, ui->xcontext) ?
                   "direct" : "indirect");

      // Initialize buffer swaps synchronized with the VBlank interrupt.
      if (ui->flags & FLAG_GLX_SWAP_CONTROL)
        ui->glx.swap_interval (1);

      const gchar* version = (const gchar*) glGetString (GL_VERSION);
      const gchar* vendor = (const gchar*) glGetString (GL_VENDOR);
      const gchar* renderer = (const gchar*) glGetString (GL_RENDERER);
      const gchar* extensions = (const gchar*) glGetString (GL_EXTENSIONS);
      DEFACE_LOG (DEFACE_INFO, "OpenGL version: %s", version);
      DEFACE_LOG (DEFACE_INFO, "OpenGL vendor: %s", vendor);
      DEFACE_LOG (DEFACE_INFO, "OpenGL renderer: %s", renderer);

      // Search for supported OpenGL extensions.
      const struct { const gchar* name; const guint16 flag; } extension[] = {
        { "GL_ARB_fragment_program", FLAG_OPENGL_ARBFP },
        { "GL_ARB_vertex_buffer_object", FLAG_OPENGL_VBO },
        { "GL_ARB_pixel_buffer_object", FLAG_OPENGL_PBO },
        { "GL_ARB_texture_non_power_of_two", FLAG_OPENGL_NPOT },
        { NULL, 0 }
      };
      guint16 flags = 0;
      for (gint32 i = 0; extension[i].name != NULL; i++)
        if (is_extension_supported (extensions, extension[i].name) == TRUE)
          flags |= extension[i].flag;

      // Load mandatory fragment-program extension functions. The presence
      // of this extension ensures OpenGL version is at least 1.3.
      if (flags & FLAG_OPENGL_ARBFP) {
        OpenglExtensions* gl = &ui->gl;
        gl->gen_programs = (GenProgramsFunc) glXGetProcAddressARB (
            (const GLubyte*) "glGenProgramsARB");
        gl->delete_programs = (DeleteProgramsFunc) glXGetProcAddressARB (
            (const GLubyte*) "glDeleteProgramsARB");
        gl->program_string = (ProgramStringFunc) glXGetProcAddressARB (
            (const GLubyte*) "glProgramStringARB");
        gl->bind_program = (BindProgramFunc) glXGetProcAddressARB (
            (const GLubyte*) "glBindProgramARB");
        gl->program_local_parameter_4fv =
            (ProgramLocalParameter4fvFunc) glXGetProcAddressARB (
                (const GLubyte*) "glProgramLocalParameter4fvARB");
        gl->get_program_iv = (GetProgramivFunc) glXGetProcAddressARB (
            (const GLubyte*) "glGetProgramivARB");
        if (gl->gen_programs != NULL && gl->delete_programs != NULL &&
            gl->program_string != NULL && gl->bind_program != NULL &&
            gl->program_local_parameter_4fv != NULL &&
            gl->get_program_iv != NULL) {

          // Load GL_ARB_vertex_buffer_object and GL_ARB_pixel_buffer_object
          // extension functions (these extensions share the same symbols).
          if (flags & FLAG_OPENGL_VBO) {
            gl->bind_buffer = (BindBufferFunc) glXGetProcAddressARB (
                (const GLubyte*) "glBindBufferARB");
            gl->delete_buffers = (DeleteBuffersFunc) glXGetProcAddressARB (
                (const GLubyte*) "glDeleteBuffersARB");
            gl->gen_buffers = (GenBuffersFunc) glXGetProcAddressARB (
                (const GLubyte*) "glGenBuffersARB");
            gl->buffer_data = (BufferDataFunc) glXGetProcAddressARB (
                (const GLubyte*) "glBufferDataARB");
            gl->buffer_sub_data = (BufferSubDataFunc) glXGetProcAddressARB (
                (const GLubyte*) "glBufferSubDataARB");
            gl->map_buffer = (MapBufferFunc) glXGetProcAddressARB (
                (const GLubyte*) "glMapBufferARB");
            gl->unmap_buffer = (UnmapBufferFunc) glXGetProcAddressARB (
                (const GLubyte*) "glUnmapBufferARB");
            if (gl->bind_buffer == NULL || gl->delete_buffers == NULL ||
                gl->gen_buffers == NULL || gl->buffer_data == NULL ||
                gl->buffer_sub_data == NULL || gl->map_buffer == NULL ||
                gl->unmap_buffer == NULL) {
              flags &= ~(FLAG_OPENGL_VBO | FLAG_OPENGL_PBO);
            }
          }

          DEFACE_LOG (DEFACE_INFO, "GL_ARB_vertex_buffer_object: %s",
                      flags & FLAG_OPENGL_VBO ? "yes" : "no");
          DEFACE_LOG (DEFACE_INFO, "GL_ARB_pixel_buffer_object: %s",
                      flags & FLAG_OPENGL_PBO ? "yes" : "no");
          DEFACE_LOG (DEFACE_INFO, "GL_ARB_texture_non_power_of_two: %s",
                      flags & FLAG_OPENGL_NPOT ? "yes" : "no");

          ui->flags |= flags;

          return TRUE;
        } else {
          deface_quit (deface, error_message[ERROR_OPENGL_ARBFP_BROKEN]);
        }
      } else {
        deface_quit (deface, error_message[ERROR_OPENGL_ARBFP_REQUIRED]);
      }
    } else {
      deface_quit (deface, error_message[ERROR_CANNOT_BIND_CONTEXT]);
    }
  } else {
    deface_quit (deface, error_message[ERROR_CANNOT_CREATE_CONTEXT]);
  }
  return FALSE;
}

static void fini_opengl (DefaceUI* ui) {
  DEFACE_DASSERT (g_thread_self () == ui->thread);
  glXMakeCurrent (ui->xdisplay, None, NULL);
  glXDestroyContext (ui->xdisplay, ui->xcontext);
}

static void fini_window (DefaceUI* ui) {
  DEFACE_DASSERT (g_thread_self () == ui->thread);
  g_object_unref (ui->window);
  XFree (ui->xvisual);
}

static void update_theme (DefaceUI* ui) {
}

static void desktop_compositing_changed_cb (GdkScreen* screen, DefaceUI* ui) {
  if (gdk_screen_is_composited (screen)) {
    DEFACE_DLOG (DEFACE_INFO, "screen compositing changed: 1");
    ui->flags |= FLAG_SCREEN_COMPOSITED;
  } else {
    DEFACE_DLOG (DEFACE_INFO, "screen compositing changed: 0");
    ui->flags &= ~FLAG_SCREEN_COMPOSITED;
  }
  update_theme (ui);
}

// FIXME(loicm): Make sure the several implementation limits are not reached.
static gboolean create_fragment_programs (DefaceUI* ui) {
  DEFACE_DASSERT (g_thread_self () == ui->thread);

  OpenglExtensions* gl = &ui->gl;
  gl->gen_programs (DEFACE_NR_PROGRAMS, ui->program_id);
  for (gint32 i = 0; i < DEFACE_NR_PROGRAMS; i++) {
    gl->bind_program (GL_FRAGMENT_PROGRAM_, ui->program_id[i]);
    gl->program_string (GL_FRAGMENT_PROGRAM_, GL_PROGRAM_FORMAT_ASCII_,
                        deface_program[i].length, deface_program[i].source);
    gint32 error_position;
    glGetIntegerv (GL_PROGRAM_ERROR_POSITION_, &error_position);
    if (error_position != -1) {
      const guchar *error_string = glGetString (GL_PROGRAM_ERROR_STRING_);
      DEFACE_LOG (DEFACE_ERROR, "program %d compilation error at "
                  "position %d:\n%.30s\n%s", i, error_position,
                  deface_program[i].source + error_position, error_string);
      return FALSE;
    }
  }
  return TRUE;
}

static gboolean create_textures (DefaceUI* ui) {
  DEFACE_DASSERT (g_thread_self () == ui->thread);
  glGenTextures (DEFACE_NR_TEXTURES, ui->texture_id);

  // Create the frame pattern texture.
  // FIXME(loicm): Only one channel is used, convert the image.
  GError *error = NULL;
  GdkPixbuf* pixbuf = gdk_pixbuf_new_from_file (
      DEFACE_DATA_DIR "/pixmaps/frame_pattern.png", &error);
  if (error == NULL) {
    glBindTexture (GL_TEXTURE_2D, ui->texture_id[DEFACE_TEXTURE_FRAME_PATTERN]);
    set_sampling_state (GL_TEXTURE_2D, GL_LINEAR, GL_LINEAR, GL_REPEAT,
                        GL_REPEAT);
    glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, gdk_pixbuf_get_width (pixbuf),
                  gdk_pixbuf_get_height (pixbuf), 0, GL_RGBA, GL_UNSIGNED_BYTE,
                  (GLvoid*) gdk_pixbuf_get_pixels (pixbuf));
    g_object_unref (pixbuf);

    return TRUE;
  } else {
    DEFACE_LOG (DEFACE_ERROR, "%s", error->message);
    return FALSE;
  }
}

static void store_theme_colors (DefaceUI* ui) {
  DEFACE_DASSERT (g_thread_self () == ui->thread);
  DefaceTheme* theme = &ui->theme;

  // FIXME(loicm): We're currently using hard-coded values waiting for correct
  //     GTK+ theme color retrieving code.
  // gdk_threads_enter ();
  // GtkSettings* settings = gtk_settings_get_for_screen (ui->screen);
  // GtkStyle* style = gtk_rc_get_style_by_paths (settings, NULL, NULL,
  //                                              GTK_TYPE_WIDGET);
  // gdk_threads_leave ();
  // theme->focused.r = (style->bg[GTK_STATE_SELECTED].red / 65535.0f) * 255;
  // theme->focused.g = (style->bg[GTK_STATE_SELECTED].green / 65535.0f) * 255;
  // theme->focused.b = (style->bg[GTK_STATE_SELECTED].blue / 65535.0f) * 255;
  // theme->unfocused.r = (style->bg[GTK_STATE_INSENSITIVE].red / 65535.0f) * 255;
  // theme->unfocused.g = (style->bg[GTK_STATE_INSENSITIVE].green / 65535.0f) * 255;
  // theme->unfocused.b = (style->bg[GTK_STATE_INSENSITIVE].blue / 65535.0f) * 255;
  theme->focused.r = 134;
  theme->focused.g = 171;
  theme->focused.b = 205;
  theme->unfocused.r = 181;
  theme->unfocused.g = 178;
  theme->unfocused.b = 172;

  DEFACE_DLOG (DEFACE_INFO, "focused color: ff%02x%02x%02x",
               theme->focused.r, theme->focused.g, theme->focused.b);
  DEFACE_DLOG (DEFACE_INFO, "unfocused color: ff%02x%02x%02x",
               theme->unfocused.r, theme->unfocused.g, theme->unfocused.b);
}

static gboolean init_ui (Deface* deface) {
  DEFACE_DASSERT (g_thread_self () == deface->ui.thread);
  DefaceUI* ui = &deface->ui;

  // Create our static resources.
  if (create_fragment_programs (ui)) {
    if (create_textures (ui)) {

      // Setup initial OpenGL states we know are gonna be fixed along the
      // application lifetime.
      glEnableClientState (GL_VERTEX_ARRAY);
      glEnableClientState (GL_COLOR_ARRAY);
      glEnableClientState (GL_TEXTURE_COORD_ARRAY);
      glDisable (GL_DEPTH_TEST);
      glEnable (GL_FRAGMENT_PROGRAM_);
#ifndef NDEBUG
      glClearColor (0.0f, 0.75f, 0.0f, 1.0f);
#else
      glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
#endif  // NDEBUG

      guint8* colors = ui->vb.colors;
      {
        // Set face area vertex colors.
        const guint32 color = PACK_UINT8 (
            (FACE_AREA_R * FACE_AREA_TRANSLUCENCY) / 255,
            (FACE_AREA_G * FACE_AREA_TRANSLUCENCY) / 255,
            (FACE_AREA_B * FACE_AREA_TRANSLUCENCY) / 255,
            FACE_AREA_TRANSLUCENCY);
        for (int i = 8; i < 12; i++)
          SET_VERTEX_COLOR_UINT32 (colors, i, color);
      }{
        // Set outer frame border vertex colors.
        const guint32 color = PACK_UINT8 (
            (FRAME_OUTER_BORDER_R * FRAME_BORDER_TRANSLUCENCY) / 255,
            (FRAME_OUTER_BORDER_G * FRAME_BORDER_TRANSLUCENCY) / 255,
            (FRAME_OUTER_BORDER_B * FRAME_BORDER_TRANSLUCENCY) / 255,
            FRAME_BORDER_TRANSLUCENCY);
        for (int i = 12; i < 28; i++)
          SET_VERTEX_COLOR_UINT32 (colors, i, color);
      }{
        // Set inner frames border vertex colors.
        const guint32 color = PACK_UINT8 (
            (FRAME_INNER_BORDER_R * FRAME_BORDER_TRANSLUCENCY) / 255,
            (FRAME_INNER_BORDER_G * FRAME_BORDER_TRANSLUCENCY) / 255,
            (FRAME_INNER_BORDER_B * FRAME_BORDER_TRANSLUCENCY) / 255,
            FRAME_BORDER_TRANSLUCENCY);
        for (int i = 28; i < 52; i++)
          SET_VERTEX_COLOR_UINT32 (colors, i, color);
      }

      guint16* indices = ui->vb.indices;
      // Set control area triangle indices.
      SET_TRIANGLE_INDICES (indices, 0, 0, 1, 2);
      SET_TRIANGLE_INDICES (indices, 1, 2, 1, 3);
      SET_TRIANGLE_INDICES (indices, 2, 1, 4, 3);
      SET_TRIANGLE_INDICES (indices, 3, 3, 4, 5);
      SET_TRIANGLE_INDICES (indices, 4, 4, 6, 5);
      SET_TRIANGLE_INDICES (indices, 5, 5, 6, 7);
      SET_TRIANGLE_INDICES (indices, 6, 6, 0, 7);
      SET_TRIANGLE_INDICES (indices, 7, 7, 0, 2);
      // Set face area triangle indices.
      SET_TRIANGLE_INDICES (indices, 8, 8, 9, 10);
      SET_TRIANGLE_INDICES (indices, 9, 10, 9, 11);
      // Set outer frame border triangle indices.
      SET_TRIANGLE_INDICES (indices, 10, 12, 13, 14);
      SET_TRIANGLE_INDICES (indices, 11, 14, 13, 15);
      SET_TRIANGLE_INDICES (indices, 12, 16, 17, 18);
      SET_TRIANGLE_INDICES (indices, 13, 18, 17, 19);
      SET_TRIANGLE_INDICES (indices, 14, 20, 21, 22);
      SET_TRIANGLE_INDICES (indices, 15, 22, 21, 23);
      SET_TRIANGLE_INDICES (indices, 16, 24, 25, 26);
      SET_TRIANGLE_INDICES (indices, 17, 26, 25, 27);
      // Set inner frame border triangle indices.
      SET_TRIANGLE_INDICES (indices, 18, 28, 29, 30);
      SET_TRIANGLE_INDICES (indices, 19, 30, 29, 31);
      SET_TRIANGLE_INDICES (indices, 20, 32, 33, 34);
      SET_TRIANGLE_INDICES (indices, 21, 34, 33, 35);
      SET_TRIANGLE_INDICES (indices, 22, 36, 37, 38);
      SET_TRIANGLE_INDICES (indices, 23, 38, 37, 39);
      SET_TRIANGLE_INDICES (indices, 24, 40, 41, 42);
      SET_TRIANGLE_INDICES (indices, 25, 42, 41, 43);
      // Set inner area border triangle indices.
      SET_TRIANGLE_INDICES (indices, 26, 44, 45, 46);
      SET_TRIANGLE_INDICES (indices, 27, 46, 45, 47);
      SET_TRIANGLE_INDICES (indices, 28, 45, 48, 47);
      SET_TRIANGLE_INDICES (indices, 29, 47, 48, 49);
      SET_TRIANGLE_INDICES (indices, 30, 48, 50, 49);
      SET_TRIANGLE_INDICES (indices, 31, 49, 50, 51);
      SET_TRIANGLE_INDICES (indices, 32, 50, 44, 51);
      SET_TRIANGLE_INDICES (indices, 33, 51, 44, 46);
      // Set upper-left corner triangle indices.
      SET_TRIANGLE_INDICES (indices, 34, 52, 53, 54);
      SET_TRIANGLE_INDICES (indices, 35, 54, 53, 55);
      // Set upper-right corner triangle indices.
      SET_TRIANGLE_INDICES (indices, 36, 56, 57, 58);
      SET_TRIANGLE_INDICES (indices, 37, 58, 57, 59);
      // Set lower-left corner triangle indices.
      SET_TRIANGLE_INDICES (indices, 38, 60, 61, 62);
      SET_TRIANGLE_INDICES (indices, 39, 62, 61, 63);
      // Set lower-right corner triangle indices.
      SET_TRIANGLE_INDICES (indices, 40, 64, 65, 66);
      SET_TRIANGLE_INDICES (indices, 41, 66, 65, 67);

      store_theme_colors (ui);
      g_signal_connect (G_OBJECT (ui->screen), "composited-changed",
                        G_CALLBACK (desktop_compositing_changed_cb), ui);

      return TRUE;
    } else {
      deface_quit (deface, error_message[ERROR_CANNOT_CREATE_TEXTURE]);
    }
  } else {
    deface_quit (deface, error_message[ERROR_CANNOT_COMPILE_PROGRAM]);
  }
  return FALSE;
}

static void update_geometry (DefaceUI* ui, const DefaceRect* geometry) {
  DEFACE_DASSERT (g_thread_self () == ui->thread);
  gfloat* positions = ui->vb.positions;
  gfloat* texcoords = ui->vb.texcoords;
  const gfloat w = geometry->width;
  const gfloat h = geometry->height;
  const gfloat border = 1.0f;
  const gfloat up = 40.0f;
  const gfloat bottom = 40.0f;
  const gfloat lr = 7.0f;

  // Set control area vertex positions.
  SET_VERTEX_POSITION (positions, 0, 0.0f, 0.0f);
  SET_VERTEX_POSITION (positions, 1, w, 0.0f);
  SET_VERTEX_POSITION (positions, 2, lr, up);
  SET_VERTEX_POSITION (positions, 3, w-lr, up);
  SET_VERTEX_POSITION (positions, 4, w, h);
  SET_VERTEX_POSITION (positions, 5, w-lr, h-bottom);
  SET_VERTEX_POSITION (positions, 6, 0.0f, h);
  SET_VERTEX_POSITION (positions, 7, lr, h-bottom);
  // Set face area vertex positions.
  SET_VERTEX_POSITION (positions, 8, lr, up);
  SET_VERTEX_POSITION (positions, 9, w-lr, up);
  SET_VERTEX_POSITION (positions, 10, lr, h-bottom);
  SET_VERTEX_POSITION (positions, 11, w-lr, h-bottom);
  // Set outer frame border vertex positions.
  SET_VERTEX_POSITION (positions, 12, CORNER_RADIUS, 0.0f);
  SET_VERTEX_POSITION (positions, 13, w-CORNER_RADIUS, 0.0f);
  SET_VERTEX_POSITION (positions, 14, CORNER_RADIUS, border);
  SET_VERTEX_POSITION (positions, 15, w-CORNER_RADIUS, border);
  SET_VERTEX_POSITION (positions, 16, CORNER_RADIUS, h-border);
  SET_VERTEX_POSITION (positions, 17, w-CORNER_RADIUS, h-border);
  SET_VERTEX_POSITION (positions, 18, CORNER_RADIUS, h);
  SET_VERTEX_POSITION (positions, 19, w-CORNER_RADIUS, h);
  SET_VERTEX_POSITION (positions, 20, 0.0f, CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 21, border, CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 22, 0.0f, h-CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 23, border, h-CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 24, w-border, CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 25, w, CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 26, w-border, h-CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 27, w, h-CORNER_RADIUS);
  // Set inner frame border vertex positions.
  SET_VERTEX_POSITION (positions, 28, CORNER_RADIUS, border);
  SET_VERTEX_POSITION (positions, 29, w-CORNER_RADIUS, border);
  SET_VERTEX_POSITION (positions, 30, CORNER_RADIUS, 2.0f*border);
  SET_VERTEX_POSITION (positions, 31, w-CORNER_RADIUS, 2.0f*border);
  SET_VERTEX_POSITION (positions, 32, CORNER_RADIUS, h-2.0f*border);
  SET_VERTEX_POSITION (positions, 33, w-CORNER_RADIUS, h-2.0f*border);
  SET_VERTEX_POSITION (positions, 34, CORNER_RADIUS, h-border);
  SET_VERTEX_POSITION (positions, 35, w-CORNER_RADIUS, h-border);
  SET_VERTEX_POSITION (positions, 36, border, CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 37, 2.0f*border, CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 38, border, h-CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 39, 2.0f*border, h-CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 40, w-2.0f*border, CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 41, w-border, CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 42, w-2.0f*border, h-CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 43, w-border, h-CORNER_RADIUS);
  // Set inner frame area border vertex positions.
  SET_VERTEX_POSITION (positions, 44, lr-border, up-border);
  SET_VERTEX_POSITION (positions, 45, w-lr+border, up-border);
  SET_VERTEX_POSITION (positions, 46, lr, up);
  SET_VERTEX_POSITION (positions, 47, w-lr, up);
  SET_VERTEX_POSITION (positions, 48, w-lr+border, h-bottom+border);
  SET_VERTEX_POSITION (positions, 49, w-lr, h-bottom);
  SET_VERTEX_POSITION (positions, 50, lr-border, h-bottom+border);
  SET_VERTEX_POSITION (positions, 51, lr, h-bottom);
  // Set upper-left rounded corner vertex positions.
  SET_VERTEX_POSITION (positions, 52, 0.0f, 0.0f);
  SET_VERTEX_POSITION (positions, 53, CORNER_RADIUS, 0.0f);
  SET_VERTEX_POSITION (positions, 54, 0.0f, CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 55, CORNER_RADIUS, CORNER_RADIUS);
  // Set upper-right rounded corner vertex positions.
  SET_VERTEX_POSITION (positions, 56, w-CORNER_RADIUS, 0.0f);
  SET_VERTEX_POSITION (positions, 57, w, 0.0f);
  SET_VERTEX_POSITION (positions, 58, w-CORNER_RADIUS, CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 59, w, CORNER_RADIUS);
  // Set lower-left rounded corner vertex positions.
  SET_VERTEX_POSITION (positions, 60, 0.0f, h-CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 61, CORNER_RADIUS, h-CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 62, 0.0f, h);
  SET_VERTEX_POSITION (positions, 63, CORNER_RADIUS, h);
  // Set lower-right rounded corner vertex positions.
  SET_VERTEX_POSITION (positions, 64, w-CORNER_RADIUS, h-CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 65, w, h-CORNER_RADIUS);
  SET_VERTEX_POSITION (positions, 66, w-CORNER_RADIUS, h);
  SET_VERTEX_POSITION (positions, 67, w, h);

  // Set control area vertex texture coordinates.
  SET_VERTEX_TEXCOORD (texcoords, 0, 0.0f, 0.0f);
  SET_VERTEX_TEXCOORD (texcoords, 1, w/w, 0.0f);
  SET_VERTEX_TEXCOORD (texcoords, 2, lr/w, up/h);
  SET_VERTEX_TEXCOORD (texcoords, 3, (w-lr)/w, up/h);
  SET_VERTEX_TEXCOORD (texcoords, 4, w/w, h/h);
  SET_VERTEX_TEXCOORD (texcoords, 5, (w-lr)/w, (h-bottom)/h);
  SET_VERTEX_TEXCOORD (texcoords, 6, 0.0f, h/h);
  SET_VERTEX_TEXCOORD (texcoords, 7, lr/w, (h-bottom)/h);
  // Set upper-left rounded corner vertex texture coordinates.
  SET_VERTEX_TEXCOORD (texcoords, 52, CORNER_RADIUS, CORNER_RADIUS);
  SET_VERTEX_TEXCOORD (texcoords, 53, 0.0f, CORNER_RADIUS);
  SET_VERTEX_TEXCOORD (texcoords, 54, CORNER_RADIUS, 0.0f);
  SET_VERTEX_TEXCOORD (texcoords, 55, 0.0f, 0.0f);
  // Set upper-right rounded corner vertex texture coordinates.
  SET_VERTEX_TEXCOORD (texcoords, 56, 0.0f, CORNER_RADIUS);
  SET_VERTEX_TEXCOORD (texcoords, 57, CORNER_RADIUS, CORNER_RADIUS);
  SET_VERTEX_TEXCOORD (texcoords, 58, 0.0f, 0.0f);
  SET_VERTEX_TEXCOORD (texcoords, 59, CORNER_RADIUS, 0.0f);
  // Set lower-left rounded corner vertex texture coordinates.
  SET_VERTEX_TEXCOORD (texcoords, 60, CORNER_RADIUS, 0.0f);
  SET_VERTEX_TEXCOORD (texcoords, 61, 0.0f, 0.0f);
  SET_VERTEX_TEXCOORD (texcoords, 62, CORNER_RADIUS, CORNER_RADIUS);
  SET_VERTEX_TEXCOORD (texcoords, 63, 0.0f, CORNER_RADIUS);
  // Set lower-right rounded corner vertex texture coordinates.
  SET_VERTEX_TEXCOORD (texcoords, 64, 0.0f, 0.0f);
  SET_VERTEX_TEXCOORD (texcoords, 65, CORNER_RADIUS, 0.0f);
  SET_VERTEX_TEXCOORD (texcoords, 66, 0.0f, CORNER_RADIUS);
  SET_VERTEX_TEXCOORD (texcoords, 67, CORNER_RADIUS, CORNER_RADIUS);

  // Store the new window geometry.
  memcpy (&ui->geometry, geometry, sizeof (DefaceRect));
}

static inline GdkEvent* unqueue_next_event (DefaceUI* ui) {
  gdk_threads_enter ();
  GdkEvent* event = gdk_display_get_event (ui->display);
  gdk_threads_leave ();
  return event;
}

static void dispatch_events (Deface* deface) {
  DEFACE_DASSERT (g_thread_self () == deface->ui.thread);

  DefaceUI* ui = &deface->ui;
  GdkEvent* event = unqueue_next_event (ui);
  while (event != NULL) {
    switch (event->type) {

      case GDK_DELETE: {
        DEFACE_DLOG (DEFACE_INFO, "GDK_DELETE");
        // Send an async request to quit the application (with no error).
        deface_quit (deface, NULL);
        break;
      }

      case GDK_KEY_PRESS: {
        GdkEventKey* key = (GdkEventKey*) event;
        DEFACE_DLOG (DEFACE_INFO, "GDK_KEY_PRESS keyval=%s is_modifier=%u",
                     gdk_keyval_name (key->keyval), key->is_modifier);
        break;
      }

      case GDK_CONFIGURE: {
        GdkEventConfigure* config = (GdkEventConfigure*) event;
        DEFACE_DLOG (DEFACE_INFO, "GDK_CONFIGURE x=%d y=%d width=%d "
                     "height=%d", config->x, config->y, config->width,
                     config->height);
        // Update the projection and viewport area. The projection matrix is
        // defined so that vertex coordinates placed at z=0 perfectly map with
        // the screen's pixel grid. That's usually achieved with an orthographic
        // projection which removes the perspective, here we keep it. Projection
        // is off-centered so that the origin is placed at the upper-left.
        const gfloat z_eye = 2.0f;
        const gfloat z_near = 1.0f;
        const gfloat z_far = -1.0f;
        const gfloat dn = z_eye - z_near;
        const gfloat df = z_eye - z_far;
        const gfloat sz = -(df + dn) / (df - dn);
        const gfloat tz = -(2.0f * dn * df) / (df - dn);
        const DefaceRect geometry = {{{
          (gfloat) config->x, (gfloat) config->y,
          (gfloat) config->width, (gfloat) config->height
        }}};
        const gfloat projection[16] = {
          (2.0f * z_eye) / geometry.width, 0.0f, 0.0f, 0.0f,
          0.0f, -(2.0f * z_eye) / geometry.height, 0.0f, 0.0f,
          0.0f, 0.0f, tz - sz * z_eye, sz,
          -z_eye, z_eye, -1.0f, z_eye
        };
        glMatrixMode (GL_PROJECTION);
        glLoadMatrixf (projection);
        glMatrixMode (GL_MODELVIEW);
        glViewport (0, 0, geometry.width, geometry.height);
        update_geometry (&deface->ui, &geometry);
        break;
      }

      case GDK_FOCUS_CHANGE: {
        GdkEventFocus* focus = (GdkEventFocus*) event;
        DEFACE_DLOG (DEFACE_INFO, "GDK_FOCUS_CHANGE in=%d", focus->in);
        DefaceTheme* theme = &ui->theme;
        const guint8 a = FRAME_TRANSLUCENCY;
        guint8 r, g, b;
        if (focus->in) {
          r = (theme->focused.r * a) / 255;
          g = (theme->focused.g * a) / 255;
          b = (theme->focused.b * a) / 255;
        } else {
          r = (theme->unfocused.r * a) / 255;
          g = (theme->unfocused.g * a) / 255;
          b = (theme->unfocused.b * a) / 255;
        }
        guint8* c = ui->vb.colors;
        // Set control area vertex colors.
        const guint32 color = PACK_UINT8 (r, g, b, a);
        for (int i = 0; i < 8; i++) {
          SET_VERTEX_COLOR_UINT32 (c, i, color);
        }
        break;
      }

      case GDK_BUTTON_PRESS: {
        GdkEventButton* button = (GdkEventButton*) event;
        DEFACE_DLOG (DEFACE_INFO, "GDK_BUTTON_PRESS x=%.2f y=%.2f button=%d "
                     "x_root=%.2f y_root=%.2f", button->x, button->y,
                     button->button, button->x_root, button->y_root);
        if (button->button == 1) {
          DEFACE_DLOG (DEFACE_INFO, "starting window move drag");
          gdk_threads_enter ();
          gdk_window_raise (ui->window);
          gdk_window_begin_move_drag (ui->window, 1, button->x_root,
                                      button->y_root, button->time);
          gdk_flush ();
          gdk_threads_leave ();
        }
        break;
      }

      case GDK_BUTTON_RELEASE: {
        GdkEventButton* button = (GdkEventButton*) event;
        DEFACE_DLOG (DEFACE_INFO, "GDK_BUTTON_RELEASE x=%.2f y=%.2f button=%d "
                     "x_root=%.2f y_root=%.2f", button->x, button->y,
                     button->button, button->x_root, button->y_root);
        break;
      }

      case GDK_MOTION_NOTIFY: {
        GdkEventMotion* motion = (GdkEventMotion*) event;
        DEFACE_DLOG (DEFACE_INFO, "GDK_MOTION_NOTIFY x=%.2f y=%.2f x_root=%.2f "
                     "y_root=%.2f", motion->x, motion->y, motion->x_root,
                     motion->y_root);
        break;
      }

      case GDK_SETTING: {
        GdkEventSetting* setting = (GdkEventSetting*) event;
        DEFACE_DLOG (DEFACE_INFO, "GDK_SETTING name=%s", setting->name);
        // FIXME(loicm): Broken detection of theme color change.
        // if (!strncmp (setting->name, "gtk-color-scheme", 16)) {
        //   store_theme_colors (ui);
        // }
        break;
      }

      default: {
        break;
      }
    }
    gdk_event_free (event);
    event = unqueue_next_event (ui);
  }
}

static void render (DefaceUI* ui) {
  DEFACE_DASSERT (g_thread_self () == ui->thread);
  OpenglExtensions* gl = &ui->gl;

  glClear (GL_COLOR_BUFFER_BIT);

  glVertexPointer (2, GL_FLOAT, 0, ui->vb.positions);
  glColorPointer (4, GL_UNSIGNED_BYTE, 0, ui->vb.colors);
  glTexCoordPointer (2, GL_FLOAT, 0, ui->vb.texcoords);

  // Draw the control area.
  glDisable (GL_BLEND);
  gl->bind_program (GL_FRAGMENT_PROGRAM_,
                    ui->program_id[DEFACE_PROGRAM_GLOSSY_WINDOW]);
  gl->program_local_parameter_4fv (GL_FRAGMENT_PROGRAM_, 0, ui->geometry.v);
  glBindTexture (GL_TEXTURE_2D,
                 ui->texture_id[DEFACE_TEXTURE_FRAME_PATTERN]);
  glDrawElements (GL_TRIANGLES, 24, GL_UNSIGNED_SHORT, &ui->vb.indices[0]);

  // Draw the face area background.
  gl->bind_program (GL_FRAGMENT_PROGRAM_,
                    ui->program_id[DEFACE_PROGRAM_FRAGMENT_COLOR]);
  glBindTexture (GL_TEXTURE_2D, 0);
  glDrawElements (GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, &ui->vb.indices[24]);

  // Draw the borders over the control area.
  glEnable (GL_BLEND);
  set_blending_state (BLENDING_SRC_OVER);
  glDrawElements (GL_TRIANGLES, 72, GL_UNSIGNED_SHORT, &ui->vb.indices[30]);

  // Draw the rounded corners. The first one masks out the frame pixels outside
  // of the arc. The second pass blends the borders over the rounded corner.
  const gfloat inner_color[4] = {
    (FRAME_INNER_BORDER_R / 255.0f) * (FRAME_BORDER_TRANSLUCENCY / 255.0f),
    (FRAME_INNER_BORDER_R / 255.0f) * (FRAME_BORDER_TRANSLUCENCY / 255.0f),
    (FRAME_INNER_BORDER_R / 255.0f) * (FRAME_BORDER_TRANSLUCENCY / 255.0f),
    FRAME_BORDER_TRANSLUCENCY / 255.0f,
  };
  const gfloat outer_color[4] = {
    (FRAME_OUTER_BORDER_R / 255.0f) * (FRAME_BORDER_TRANSLUCENCY / 255.0f),
    (FRAME_OUTER_BORDER_R / 255.0f) * (FRAME_BORDER_TRANSLUCENCY / 255.0f),
    (FRAME_OUTER_BORDER_R / 255.0f) * (FRAME_BORDER_TRANSLUCENCY / 255.0f),
    FRAME_BORDER_TRANSLUCENCY / 255.0f,
  };
  gl->bind_program (GL_FRAGMENT_PROGRAM_,
                    ui->program_id[DEFACE_PROGRAM_FILLED_ARC]);
  set_blending_state (BLENDING_DST_IN);
  glDrawElements (GL_TRIANGLES, 24, GL_UNSIGNED_SHORT, &ui->vb.indices[102]);
  gl->bind_program (GL_FRAGMENT_PROGRAM_,
                    ui->program_id[DEFACE_PROGRAM_STROKED_ARCS]);
  gl->program_local_parameter_4fv (GL_FRAGMENT_PROGRAM_, 0, inner_color);
  gl->program_local_parameter_4fv (GL_FRAGMENT_PROGRAM_, 1, outer_color);
  set_blending_state (BLENDING_SRC_OVER);
  glDrawElements (GL_TRIANGLES, 24, GL_UNSIGNED_SHORT, &ui->vb.indices[102]);

  // Buffer swap, throttling the loop waiting for the VBlank interrupt.
  if (ui->flags & FLAG_GLX_SWAP_CONTROL) {
    glXSwapBuffers (ui->xdisplay, ui->xwindow);
  } else if (ui->flags & FLAG_GLX_VIDEO_SYNC) {
    GlxExtensions* glx = &ui->glx;
    guint32 retrace_count;
    glx->get_video_sync (&retrace_count);
    glx->wait_video_sync (2, (retrace_count + 1) % 2, &retrace_count);
    glXSwapBuffers (ui->xdisplay, ui->xwindow);
  } else {
    // TODO(loicm): Fallback to manual throttling.
    glXSwapBuffers (ui->xdisplay, ui->xwindow);
  }
}

static gpointer ui_main (Deface* deface) {
  DEFACE_DLOG (DEFACE_INFO, "entered UI thread");
  DefaceUI* ui = &deface->ui;

#ifndef NDEBUG
  // Ensure the thread field is correctly set so that the DASSERTs in functions
  // called from that thread aren't mistakenly catched. The issue being that at
  // that time the return value of g_thread_create() could've not been set yet.
  ui->thread = g_thread_self ();
#endif  // NDEBUG

  // Setup X11 and OpenGL stuff.
  if (init_x11 (deface)) {
    if (init_window (deface)) {
      if (init_opengl (deface)) {
        if (init_ui (deface)) {
          update_theme (ui);
          gdk_threads_enter ();
          gdk_window_show (ui->window);
          gdk_flush ();
          gdk_threads_leave ();

          // UI main loop.
          while (!g_atomic_int_get (&ui->quit)) {
            dispatch_events (deface);
            render (ui);
          }

          // Clean up.
          gdk_threads_enter ();
          gdk_window_hide (ui->window);
          gdk_flush ();
          gdk_threads_leave ();
        }
        fini_opengl (ui);
      }
      fini_window (ui);
    }
  }

  DEFACE_DLOG (DEFACE_INFO, "leaving UI thread");
  return NULL;
}

gboolean deface_ui_spawn_thread (Deface* deface) {
  GThread* thread;
  GError* error;
  thread = g_thread_create ((GThreadFunc) ui_main, deface, TRUE, &error);
  if (thread != NULL) {
    deface->ui.thread = thread;
    return TRUE;
  } else {
    DEFACE_LOG (DEFACE_ERROR, "can't spawn UI thread: %s", error->message);
    return FALSE;
  }
}

void deface_ui_join_thread (Deface* deface) {
  DEFACE_DASSERT (g_thread_self () != deface->ui.thread);
  if (deface->ui.thread != NULL) {
    g_atomic_int_set (&deface->ui.quit, 1);
    DEFACE_DLOG (DEFACE_INFO, "joining UI thread");
    g_thread_join (deface->ui.thread);
    deface->ui.thread = NULL;
    DEFACE_DLOG (DEFACE_INFO, "joined UI thread");
  }
}
